@@ -1,5 +1,8 @@
 package Complete::Util;
 
+our $DATE = '2014-07-13'; # DATE
+our $VERSION = '0.13'; # VERSION
+
 use 5.010001;
 use strict;
 use warnings;
@@ -14,9 +17,6 @@ our @EXPORT_OK = qw(
                        complete_program
                );
 
-our $VERSION = '0.12'; # VERSION
-our $DATE = '2014-07-02'; # DATE
-
 our %SPEC;
 
 $SPEC{complete_array_elem} = {
@@ -141,11 +141,30 @@ $SPEC{complete_file} = {
     v => 1.1,
     summary => 'Complete file and directory from local filesystem',
     args => {
-        word => { schema=>[str=>{default=>''}], pos=>0 },
-        file => { summary => 'Whether to include file',
-                  schema=>[bool=>{default=>1}] },
-        dir  => { summary => 'Whether to include directory',
-                  schema=>[bool=>{default=>1}] },
+        word => {
+            schema  => [str=>{default=>''}],
+            pos     => 0,
+        },
+        filter => {
+            summary => 'Only return items matching this filter',
+            description => <<'_',
+
+Filter can either be a string or a code.
+
+For string filter, you can specify a pipe-separated groups of sequences of these
+characters: f, d, r, w, x. Dash can appear anywhere in the sequence to mean
+not/negate. An example: `f` means to only show regular files, `-f` means only
+show non-regular files, `drwx` means to show only directories which are
+readable, writable, and executable (cd-able). `wf|wd` means writable regular
+files or writable directories.
+
+For code filter, you supply a coderef. The coderef will be called for each item
+with these arguments: `$name`. It should return true if it wants the item to be
+included.
+
+_
+            schema  => ['any*' => {of => ['str*', 'code*']}],
+        },
     },
     result_naked => 1,
     result => {
@@ -153,10 +172,36 @@ $SPEC{complete_file} = {
     },
 };
 sub complete_file {
-    my %args  = @_;
-    my $word  = $args{word} // "";
-    my $f     = $args{file} // 1;
-    my $d     = $args{dir} // 1;
+    my %args   = @_;
+    my $word   = $args{word} // "";
+    my $filter = $args{filter};
+
+    if ($filter && !ref($filter)) {
+        my @seqs = split /\s*\|\s*/, $filter;
+        $filter = sub {
+            my $name = shift;
+            my @st = stat($name) or return 0;
+            my $mode = $st[2];
+            my $pass;
+          SEQ:
+            for my $seq (@seqs) {
+                my $neg = sub { $_[0] };
+                for my $c (split //, $seq) {
+                    if    ($c eq '-') { $neg = sub { $_[0] ? 0 : 1 } }
+                    elsif ($c eq 'r') { next SEQ unless $neg->($mode & 0400) }
+                    elsif ($c eq 'w') { next SEQ unless $neg->($mode & 0200) }
+                    elsif ($c eq 'x') { next SEQ unless $neg->($mode & 0100) }
+                    elsif ($c eq 'f') { next SEQ unless $neg->($mode & 0100000)}
+                    elsif ($c eq 'd') { next SEQ unless $neg->($mode & 0040000)}
+                    else {
+                        die "Unknown character in filter: $c (in $seq)";
+                    }
+                }
+                $pass = 1; last SEQ;
+            }
+            $pass;
+        };
+    }
 
     my @all;
     if ($word =~ m!(\A|/)\z!) {
@@ -175,9 +220,8 @@ sub complete_file {
 
     my @words;
     for (@all) {
-        next if (-f $_) && !$f;
-        next if (-d _ ) && !$d;
-        $_ = "$_/" if (-d _) && !m!/\z!;
+        next if $filter && !$filter->($_);
+        $_ = "$_/" if (-d $_) && !m!/\z!; # XXX double stat
         #s!.+/(.+)!$1!;
         push @words, $_;
     }
@@ -204,7 +248,7 @@ Complete::Util - General completion routines
 
 =head1 VERSION
 
-This document describes version 0.12 of Complete::Util (from Perl distribution Complete-Util), released on 2014-07-02.
+This document describes version 0.13 of Complete::Util (from Perl distribution Complete-Util), released on 2014-07-13.
 
 =head1 DESCRIPTION
 
@@ -258,13 +302,22 @@ Arguments ('*' denotes required arguments):
 
 =over 4
 
-=item * B<dir> => I<bool> (default: 1)
+=item * B<filter> => I<code|str>
 
-Whether to include directory.
+Only return items matching this filter.
 
-=item * B<file> => I<bool> (default: 1)
+Filter can either be a string or a code.
 
-Whether to include file.
+For string filter, you can specify a pipe-separated groups of sequences of these
+characters: f, d, r, w, x. Dash can appear anywhere in the sequence to mean
+not/negate. An example: C<f> means to only show regular files, C<-f> means only
+show non-regular files, C<drwx> means to show only directories which are
+readable, writable, and executable (cd-able). C<wf|wd> means writable regular
+files or writable directories.
+
+For code filter, you supply a coderef. The coderef will be called for each item
+with these arguments: C<$name>. It should return true if it wants the item to be
+included.
 
 =item * B<word> => I<str> (default: "")
 
@@ -325,7 +378,7 @@ Please visit the project's homepage at L<https://metacpan.org/release/Complete-U
 
 =head1 SOURCE
 
-Source repository is at L<https://github.com/sharyanto/perl-Complete-Util>.
+Source repository is at L<https://github.com/sharyanto/perl-SHARYANTO-Complete-Util>.
 
 =head1 BUGS
 
